#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['fastpf', 'fastpf.data', 'fastpf.grid_reduction', 'fastpf.plotting']

package_data = \
{'': ['*'],
 'fastpf': ['.pytest_cache/*', '.pytest_cache/v/*', '.pytest_cache/v/cache/*']}

install_requires = \
['loguru >= 0.6.0',
 'numpy >= 1.16.0',
 'scipy >= 1.8.0',
 'numba >= 0.55.1',
 'pathos >= 0.2.8',
 'threadpoolctl >= 3.1.0']

setup(name='fastpf',
      version='0.0.5',
      description='Performance-first Power Flow implemented in Python with AOT Compilation',
      author=None,
      author_email='Christoph Kattmann <christoph.kattmann@gmail.com>',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
     )
